#ifndef ARGS_H
#define ARGS_H
/***************************************************************************
 *   Copyright (C) 2004 - 2006 Symbian Software Ltd.                       *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "state_utils.h"
#include "file_tree.h"
#include "ptr_vector.h"
#include "get_options.h"


/*!\ingroup args_component args_component_interface
 *\file args.h
 * This file provides the Args component interface
*/

/*!\addtogroup args_interface*/
/*@{*/

/*! Enumeration of policies for discarding lines*/
typedef enum {
    DISCARD_DROP,	/*!< Drop discarded lines*/
    DISCARD_BLANK,	/*!< Blank discarded lines*/
    DISCARD_COMMENT /*!< Comment discarded lines*/
} discard_policy_t;

/*! Sequential symbolic constants for coan commands */
enum command_codes {
    CMD_HELP = 1,	/*!< The help command*/
    CMD_VERSION,	/*!< The version command*/
    CMD_SOURCE,	/*!< The source command*/
    CMD_SYMBOLS,	/*!< The symbols command*/
    CMD_INCLUDES,	/*!< The includes command*/
    CMD_DEFS,	/*!< The defs command*/
    CMD_PRAGMAS,	/*!< The pragmas command*/
    CMD_ERRORS,	/*!< The errors command*/
    CMD_DIRECTIVES	/*!< The directives command*/
};

/*! Bitwise symbolic constants for coan commands*/
enum command_flags {
    CMD_HELP_MASK = 1 << (CMD_HELP - 1),	/*!< The help command*/
    CMD_VERSION_MASK = 1 << (CMD_VERSION - 1),	/*!< The version command*/
    CMD_SOURCE_MASK = 1 << (CMD_SOURCE - 1),	/*!< The source command*/
    CMD_SYMBOLS_MASK = 1 << (CMD_SYMBOLS - 1),	/*!< The symbols command*/
    CMD_INCLUDES_MASK = 1 << (CMD_INCLUDES - 1),	/*!< The includes command*/
    CMD_DEFS_MASK = 1 << (CMD_DEFS - 1),	/*!< The defs command*/
    CMD_PRAGMAS_MASK = 1 << (CMD_PRAGMAS - 1),	/*!< The pragmas command*/
    CMD_ERRORS_MASK = 1 << (CMD_ERRORS - 1),	/*!< The pragmas command*/
    CMD_DIRECTIVES_MASK = CMD_INCLUDES_MASK | CMD_DEFS_MASK | CMD_PRAGMAS_MASK
    | CMD_ERRORS_MASK
    /*!< The directives command */
};

/*! \fn void parse_executable(char **argv);
 * Parse the full and short names of the executable.
    \param argv		The array of commandline arguments.
 *
    Element 0 of \em argv is parsed to find the complete
    pathname and filename of our executable and the addresses
    of these strings are stored respectively at \c exec_path and
    \c prog_name in the component's state.
*/
extern void
parse_executable(char **argv);

/*! Parse the commandline.
    \param argc		The number of commandline arguments.
    \param	argv	Array of commandline arguments.
 *
    Populate the component's state by parsing the commandline arguments.
*/
extern void
parse_args(int argc, char *argv[]);

/*! Analyse the component's state after parsing the commandline,
    diasgnose any errors and draw final conclusions.
*/
extern void
finish_args(void);

/*@}*/

/*!\ingroup args_interface_state_utils*/
/*@{*/

/*! The public state of the Args component
*/
PUBLIC_STATE_DEF(args)
{
    cmd_option_t * command;
    /*!< Pointer to the details of the operative coan command */
    char	*exec_path;
    /*!< argv[0]*/
    char	*prog_name;
    /*!< Filename element of \c exec_path*/
    char	*backup_suffix;
    /*!< Suffix for backup files*/
    bool	got_opts;
    /*!< Are we finished parsing the commandline?*/
    bool	replace;
    /*!< Do we replace input files with output files?*/
    bool	list_locate;
    /*!< Do we report file and line numbers for listed items?*/
    bool	list_only_once;
    /*!< Do we report only the first occurrence of listed items?*/
    bool	list_only_active;
    /*!< Do we list items only from kept lines?*/
    bool	list_only_inactive;
    /*!< Do list items only from dropped lines?*/
    bool	list_symbols_in_ifs;
    /*!< Do we list symbols in hash-if/else/endif directives?*/
    bool	list_symbols_in_defs;
    /*!< Do we list symbols in hash-define directives?*/
    bool	list_symbols_in_undefs;
    /*!< Do we list symbols in hash-undef directives?*/
    bool	list_symbols_in_includes;
    /*!< Do we list symbols in hash-include directives?*/
    bool	list_system_includes;
    /*!< Do we list system hash-include directives?*/
    bool	list_local_includes;
    /*!< Do we list local hash-include directives?*/
    bool	complement;
    /*!< Are to output lines instead of dropping tem and vice versa?*/
    bool	eval_consts;
    /*!< Do we evaluate constants in truth-functional contexts or treat them as
    	unknowns? */
    bool	resolve_symbols;
    /*!< Do we report the resolved values of reported symbols */
    discard_policy_t	discard_policy;
    /*!< Policy for discarding lines*/
    bool	line_directives;
    /*!< Do we output hash-line directives?*/
    bool	plaintext;
    /*!< Are we to omit parsing for comments?*/
    bool	recurse;
    /*!< Recurse into directories?*/
    bool	keepgoing;
    /*!< Continue to process input files after errors*/
    bool	implicit;
    /*!< Do we implicitly --undef all unconfigured symbols? */
    int		diagnostic_filter;
    /*!< Bitmask of diagnostic filters*/
}
PUBLIC_STATE_T(args);

IMPORT(args);
/*@}*/


#endif /* EOF*/
