#ifndef CANONICAL_STRING_H
#define CANONICAL_STRING_H
/***************************************************************************
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "bool.h"
#include <stddef.h>

/*!\ingroup canonical_string_component canonical_string_interface
	\file canonical_string.h
	This file provides the Canonical Line component interface.
*/

/*!	\addtogroup canonical_string_interface*/
/*@{*/

/*! Abstract type of canonical string object */
typedef struct canonical_string * canonical_string_h;

/*! Abstract type of constant canonical string object */
typedef struct canonical_string const * canonical_string_const_h;

/*! Construct a new canonical string from given text.
	The constructed canonical string contains the
	given text with all comments stripped, all
	line-continuations joined and all whitespace
	sequences condensed to a single space.

	\param	cpp	On entry, NULL, or else a pointer to the text to be scanned into a canonical string.
            On return, if not NULL on entry, stores the address of the first character not scanned;
            otherwise unchanged. if NULL, \e cs is initialised as the empty string.
    \param  nbytes  0, or the maximum number of bytes that may consumed from \e *cpp, when \e cpp is not NULL.
        If 0 when \e cpp != NULL then \c strlen(cp) is assumed.
*/
extern canonical_string_h
canonical_string_new(char const **cpp, size_t nbytes);

struct int_spec; /*!< Forward decl */

/*! Construct a new canonical string to represent a given integer as a numeral.

	\param	int_spec	An \em int_spec representing the integer to
        be returned as a string.
		string.
	\return A canonical string containing the decimal numeral
	that represents \e val/
*/
extern canonical_string_h
canonical_numeral(struct int_spec const * int_spec);

/*! Construct a new canonical string to contain a symbol name.
    \param  cpp  On entry, pointer to the address of the text to be parsed as a symbol name.
        On return, stores the address of the first scanned character that is not part of
        an identifier.
    \return A canonical string containing the parsed symbol.
*/
extern canonical_string_h
canonical_identifier(char const **cpp);


/*! Construct a copy of a canonical string.*/
extern canonical_string_h
canonical_string_copy(canonical_string_const_h src);

/*! Destroy a canonical string */
extern void
canonical_string_dispose(canonical_string_h cs);

/*! Swap two canonical strings */
extern void
canonical_string_swap(canonical_string_h lhs, canonical_string_h rhs);

/*! Assign one canonical string to another */
extern void
canonical_string_assign(canonical_string_h dest, canonical_string_const_h src);

/*! Say whether two canonical strings are equal */
extern bool
canonical_string_equal(canonical_string_const_h lhs, canonical_string_const_h rhs);

/*! Get a pointer to the text of a canonical string
	\param	cs	The canonical string to query.
	\return	A pointer to the text of \e cs.
*/
extern char const *
canonical_string_text(canonical_string_const_h cs);

/*! Get the length of a canonical string
	\param	cs	The canonical string to query.
	\return	The length of \e cs.
*/
extern size_t
canonical_string_length(canonical_string_const_h cs);


/*! Compare a canonical string with another canonical string or region of text.
	\param	lhs	First canonical string to be compared.
	\param	rhs	Second canonical string or pointer to text.
	\param	rhslen	If 0 then \e rhs addresses a second canonical string to be
			compared with \e lhs. Otherwise \e rhs addresses text
			of length \e keylen to be compared with the text of \e lhs.
	\return A negative value if \e lhs is less than \e rhs; 0 if \e lhs
		is equal to \e rhs; a positive value if \e lhs is greater than \e rhs.

*/
extern int
canonical_string_compare(canonical_string_const_h lhs,
                         void const *rhs, size_t rhslen);

/*! Append one canonical string to another
	\param	cs	The canonical string to be extended.
	\param	more	The canonical string to be appended.
	\param	punct	If true then a punctuating space is appended
			to \e cs before \e more is appended.
*/
extern void
canonical_string_append(canonical_string_h cs, canonical_string_const_h more,
                        bool punct);


/*@}*/

#endif /* EOF */
