#ifndef CHEW_H
#define CHEW_H

/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2002, 2003 Tony Finch <dot@dotat.at>.                   *
 *   All rights reserved.                                                  *
 *   Copyright (C) 1985, 1993 The Regents of the University of California. *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed by Mike Kinghan, imk@strudl.org, derived from the code    *
 *   of Tony Finch                                                         *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "state_utils.h"

/*!\ingroup chew_component chew_interface
 *\file chew.h
 * This file provides the Chew component interface
*/

/*! \addtogroup chew_interface*/
/*@{*/

/*! Enumeration of comment states*/
typedef enum comment_state {
    /*! Outside a comment*/
    NO_COMMENT = false,
    /*! In a C-style comment*/
    C_COMMENT,
    /*! In a C++-style comment, i.e. // and end of line*/
    CXX_COMMENT,
    /*! Just after forward-slash + back-slash + newline*/
    STARTING_COMMENT,
    /*! Just after star + back-slash + newline in a C comment*/
    FINISHING_COMMENT,
    /*! In the text of a hash-error directive or the definiens of a hash-define*/
    PSEUDO_COMMENT
} comment_state_t;

/*! Enumeration of states of a source line*/
typedef enum {
    /*! Only whitespace and comments so far on the line*/
    LS_NEUTER,
    /*! Only whitespace, comments, and a hash- so far on the line*/
    LS_DIRECTIVE,
    /*! Source code found on the line (text other than whitespace,
    	comments, or a <tt>#</tt>-directive)*/
    LS_CODE
} line_state_t;

/*! Say whether a <tt>char *</tt> address a line end,
	either Unix type or Windows type.
	\param	cp	Text Pointer
	\return 0 if \e cp does not address a newline sequence,
	otherwise the length of the newline sequence, i,e,
	1 for Unix '\n', 2 for Windows "\r\n"
*/
extern size_t
eol(char const *cp);

/*! Consume whitespace, comments, quotation or the text of a free-form
	directive (hash-error,hash-pragma) in the source text and stop at
	the first character beyond all these.
 *
    \param	cp	The current text pointer
    \return A pointer to the next input character that is not to be chewed.
 *
    The function replenishes the line buffer and copes with line-continuations.
 * It maintains state with respect to all syntactically significant categories
 * of context.
*/
extern char const *
chew_on(char const *cp);

/*! Consume an identifier in the source text and stop at the next character.
 *
    \param	cp	The current text pointer
    \return A pointer to the next input character that cannot be part of
    an identifier.
*/
extern char const *
chew_symbol(char const *cp);

/*! Consume whitespace in the source text and stop at the next character.
 *
    \param	cp	The current text pointer
    \return A pointer to the next non-whitespace input character.
*/
extern char const *
chew_space(char const *cp);

/*! Consume line-continuations at the input text pointer.

	\param	cp	The current text pointer
	\return Pointer to the next character of input following any
		sequence of 1 or more line-continuations, i.e. "\\n".
	If there is no such sequence \e cp	is returned unchanged.

	The line buffer is extended with each continuation by
	appending the next physical line of input.

*/
extern char const *
chew_continuation(char const *cp);


/*! Consume a preprocessor macro invocation.

    \param	cp	The current text pointer addressing a macro invocation.
    \return A pointer to the next input character that cannot be part of
    a macro invocation, or NUL at end of line.

	The function does not insist that \e cp initially addresses an
	identifier. It may also address '(' and in that case the function
	will consume an argument list until it passes a balancing ')' or
	reaches a character that cannot belong to an argument list.
*/
extern char const *
chew_macro_call(char const *cp);

/*! Consume a header file name in source text until a delimiter or end of line
	is found.

    \param	cp			The current text pointer addressing a header file name.
    \return A pointer to the next input character that cannot be part of the
		header file name, or NUL at end of line.
*/
extern char const *
chew_header_name(char const *cp);


/*! Consume a string until whitespace, comments or
	quotation are encountered, or a quoted string until it is closed.

    \param	cp	The current text pointer
    \return A pointer to the character just passed the
		consumed string.

	If \e *cp == '\"' on entry a quoted string is recognised
	and consumed until closed.

	The function consumes line-continuations.

*/
extern char const *
chew_string(char const *cp);

/*! The function behaves like \e chew_string(), but stops
	if a line-continuation is encountered.
*/
extern char const *
chew_unbroken_string(char const *cp);

/*! Reinitialise the line state and comment state*/
extern void
chew_toplevel(void);

/*@}*/

/*! \ingroup chew_interface_state_utils*/
/*@{*/

/*! The public state of the Chew component*/
PUBLIC_STATE_DEF(chew)
{
    enum comment_state comment_state;
    /*!< The current comment state*/
    bool stop_at_quote;
    /*!< Do we stop at quotes? */
    bool must_balance_quotes;
    /*!< Do we have to balance quotation marks? */
    bool in_double_quote; /*!< Are we reading within double quotes?*/
    bool in_source; /*!< Are we reading input source (as opposed to
		the configured definition of a \c -D symbol) ? */
    line_state_t line_state;
    /*!< The current line state*/
    size_t	last_quote_start_line;
    /*!< Line number of the most recent open-quote*/
    size_t	 last_comment_start_line;
    /*!< Line number of most recent open-comment*/
}
PUBLIC_STATE_T(chew);
/*@}*/

/*! \addtogroup chew_interface_state_utils*/
/*@{*/
IMPORT(chew);
/*@}*/

#endif /* EOF*/
