#ifndef EVAL_RESULT_H
#define EVAL_RESULT_H

/***************************************************************************
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Originally contributed by Mike Kinghan, imk@strudl.org.               *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "opts.h"
#include "canonical_string.h"
#include "integer_constant.h"

/*!\ingroup eval_result_component eval_result_interface
*  \file eval_result.h
   This file defines a structure to represent the result of
   evaluating an expression, and symbolic constants to represent
   attributes of such results.
*/

/*! \addtogroup eval_result_interface*/
/*@{*/

/*! Bit flags denoting findings of evaluation*/
enum eval_flags {
    EVAL_PENDING,	/*!< Expression not evaluated*/
    EVAL_SEEN = 1,
    /*!< The text has been seen by the parser.
    	This flag lets us head off infinite looping on
    	circular macro definitions*/
    EVAL_INSOLUBLE = 2,
    /*!< Evaluated text cannot be resolved*/
    EVAL_CONST = 4,
    /*!< Expression is a constant*/
    EVAL_KEEP =  8,
    /*!< Expression is not to be eliminated*/
    EVAL_KEEP_CONST = EVAL_CONST | EVAL_KEEP,
    /*!< Expression is constant not be eliminated*/
    EVAL_DEL_PAREN = 16,
    /*!< Redundant parentheses to deleted from expression*/
    EVAL_TRUE = 32,
    /*!< Expression evaluates as true*/
    EVAL_FALSE = 64,
    /*!< Expression evaluates as false*/
    EVAL_ILLEGAL = 128,
    /*!< Text is not an expression */
    EVAL_RESOLVED = EVAL_TRUE | EVAL_FALSE,
    /*!< Expression is resolved*/
    EVAL_PARSED = EVAL_RESOLVED | EVAL_KEEP,
    /*!< Expression has been fully parsed */

};

/*! Structure representing the evaluation of
	a text item */
typedef struct eval_result {
    char * sym_name;
    /*!< Name of symbol, if text is a symbol or a
    	unary function of one*/
    canonical_string_h sym_def;
    /*!< Definition of symbol, if text assigns a
    	definition to a symbol*/
    canonical_string_h most_resolved;
    /*!< Most resolved form of text. If text is a symbol this
    	is the most resolved form of its definition*/
    int_spec_t value;
    /*!< The unsigned integer value of text if it is a soluble
    	expression */
    int flags;
    /*!< Bit set of \em eval_flags*/
} eval_result_t;

/*! Static initialiser for type \e eval_result_t */
#define EVAL_RESULT_INITOR {0,0,0,INT_SPEC_INITOR,0}

/*! Set flags in an \em eval_result_t
*/
#define SET_FLAGS(eval_result,bits)	((eval_result).flags |= (bits))

/*! Set flags in an \em eval_result_t if a condition is true
*/
#define SET_FLAGS_IF(cond,eval_result,bits)\
	(void)((cond) ? SET_FLAGS(eval_result,bits) : 0)

/*! Clear flags in an \em eval_result_t
*/
#define CLEAR_FLAGS(eval_result,bits) ((eval_result).flags &= ~(bits))

/*! Clear flags in an \em eval_result_t if a condition is true
*/
#define CLEAR_FLAGS_IF(cond,eval_result,bits)\
	(void)((cond) ? CLEAR_FLAGS(eval_result,bits) : 0)

/*! Invert the values of specified flags in an
	\em eval_result_t
*/
#define FLIP_FLAGS(eval_result,bits) ((eval_result).flags ^= (bits))

/*!	 Get the flags in \em eval_result_t that match a mask*/
#define FLAGS_IN(eval_result,bits)	((eval_result).flags & (bits))

/*! Test whether specified flags are set in an \em eval_result_t
*/
#define AFFIRM_FLAGS(eval_result,bits)\
	(((eval_result).flags & (bits)) == (bits))

/*! Test whether specified flags are clear in an
	\em eval_result_t
*/
#define DENY_FLAGS(eval_result,bits)\
	(((eval_result).flags & (bits)) == 0)

/*! Test whether an \em eval_result_t
	represents a true expression
*/
#define IS_TRUE(eval_result)	AFFIRM_FLAGS(eval_result,EVAL_TRUE)

/*! Test whether an \em eval_result_t represents a false
	expression.
*/
#define IS_FALSE(eval_result)	AFFIRM_FLAGS(eval_result,EVAL_FALSE)

/*! Test whether an \em eval_result_t represents a constant.
*/
#define IS_CONST(eval_result)	AFFIRM_FLAGS(eval_result,EVAL_CONST)

/*! Return the integer width of the value of an eval result
*/
#define INT_WIDTH(eval_result)	((eval_result).value.width)

/*! Test whether an \em eval_result_t represents
	unresolved text.
*/
#define UNRESOLVED(eval_result)	DENY_FLAGS(eval_result,EVAL_FALSE | EVAL_TRUE)

/*! Test whether an \em eval_result_t represents a
	resolved expression.
*/
#define RESOLVED(eval_result)	!UNRESOLVED(eval_result)

/*! Test whether an \em eval_result_t represents
	insoluble text.
*/
#define INSOLUBLE(eval_result)	AFFIRM_FLAGS(eval_result,EVAL_INSOLUBLE)

/*! Flag an \em eval_result_t as not being a soluble expression.
*/
#define SET_INSOLUBLE(eval_result)	SET_FLAGS(eval_result,EVAL_INSOLUBLE)


/*! Test whether an \em eval_result_t represents
	a symbol already visited for evaluation.

	A circularly defined symbol can be detected when it is
	found to have been seen but flagged as \c RESOLVED or \c KEEP.
*/
#define SEEN(eval_result)	AFFIRM_FLAGS(eval_result,EVAL_SEEN)

/*! Test whether an \em eval_result_t stipulates
	deleting superfluous parentheses
*/
#define	DEL_PAREN(eval_result)	AFFIRM_FLAGS(eval_result,EVAL_DEL_PAREN)

/*! Flag an \em eval_result_t for deletion of superflous
	parentheses.
*/
#define SET_DEL_PAREN(eval_result)	SET_FLAGS(eval_result,EVAL_DEL_PAREN)

/*! Test whether an \em eval_result_t represents
	an expression we can't eliminate.
*/
#define KEEP(eval_result)	AFFIRM_FLAGS(eval_result,EVAL_KEEP)

/*! Flag an \em eval_result_t as an expression we can't
	eliminate.
*/
#define SET_KEEP(eval_result)	SET_FLAGS(eval_result,EVAL_KEEP)

/*! Test whether an \em eval_result_t represents a non-eliminable constant
*/
#define KEEP_CONST(eval_result)	AFFIRM_FLAGS(eval_result,EVAL_KEEP_CONST)

/*! Flag an \em eval_result_t as a non-eliminable constant*/
#define SET_KEEP_CONST(eval_result)	SET_FLAGS(eval_result,EVAL_KEEP_CONST)

/*! Flag an \em eval_result_t as a constant*/
#define SET_CONST(eval_result)	SET_FLAGS(eval_result,EVAL_CONST)

/*! Flag an \em eval_result_t as a non-expression */
#define SET_ILLEGAL(eval_result)	SET_FLAGS(eval_result,EVAL_ILLEGAL)

/*! Flag an \em eval_result_t as a legal expression */
#define SET_LEGAL(eval_result)	CLEAR_FLAGS(eval_result,EVAL_ILLEGAL)

/*! Test whether an \em eval_result_t represents a non-expression */
#define ILLEGAL(eval_result)	AFFIRM_FLAGS(eval_result,EVAL_ILLEGAL)

/*! Test whether an \em eval_result_t represents a legal expression */
#define LEGAL(eval_result)	!ILLEGAL(eval_result)

/*! Flag an \e eval_result_t as unresolved */
#define SET_UNRESOLVED(eval_result) \
	CLEAR_FLAGS(eval_result,EVAL_TRUE | EVAL_FALSE)

/*! Constant \em int_spec_t representing falsity */
extern int_spec_t const int_spec_false;

/*! Constant \em int_spec_t representing truth */
extern int_spec_t const int_spec_true;

/*! Assign a value in an \em eval_result_t
    \param  result   Pointer the the \em eval_result_t to which a value is to be assigned.
    \param  int_spec    Descriptor of the integer value to be assigned.
*/
extern void
eval_result_set_value(eval_result_t *result, int_spec_t const * int_spec);

/*! Zero all fields in an \e eval_result_t
	\param	result If not NULL, pointer to an \e eval_result_t to
		be cleared.

	If \e result is NULL the function is a no-op.
*/
extern void
eval_result_clear(eval_result_t *result);

/*! Say whether an \e eval_result_t has been fully parsed.
	\param	result	Pointer the \e eval_result_t to be tested.
	\return	Return true if \e result satisfies either of the flags
	\c EVAL_RESOLVED or \c EVAL_KEEP.
*/
extern bool
eval_result_parsed(eval_result_t *result);


/*@}*/

#endif /* EOF*/
