/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2002, 2003 Tony Finch <dot@dotat.at>.                   *
 *   All rights reserved.                                                  *
 *   Copyright (C) 1985, 1993 The Regents of the University of California. *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed by Mike Kinghan, imk@strudl.org, derived from the code    *
 *   of Tony Finch                                                         *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "evaluator.h"
#include "eval_result.h"
#include "io.h"
#include "report.h"
#include "line_edit.h"
#include "chew.h"
#include "args.h"
#include "categorical.h"
#include "if_control.h"
#include "configured_symbols.h"
#include "unconfigured_symbols.h"
#include "report.h"
#include "line_despatch.h"
#include "hash_include.h"
#include "includes_tally.h"
#include "line_select.h"
#include "directives_tally.h"
#include "lexicon.h"
#include "integer_ops.h"

/*!\ingroup evaluator_component evaluator_internals evaluator_interface
 *
 * \file evaluator.c
    This file implements the Evaluator component
*/

/*! \ingroup evaluator_internals_state_utils*/
/*@{*/
/*! The global state of the Evaluator component*/
STATE_DEF(evaluator)
{
    /*! Line-buffer position of the start of the current hash-if condition*/
    char const			*ifpos;
    /*! The line type of the current line*/
    line_type_t		line_type;
    /*! Levels of recursion into parsing a symbol definition.
    	We must only report symbol values for the \e --eval
    	option of the \c symbols command when this level is
    	0, i.e. for the symbol that has initiated the
    	recursion.  */
    int parsing_sym_def_depth;
}
STATE_T(evaluator);
/*@}*/

/*! \addtogroup evaluator_internals_state_utils*/
/*@{*/
NO_PUBLIC_STATE(evaluator);

IMPLEMENT(evaluator,ZERO_INITABLE);
/*@}*/

/*! \addtogroup evaluator_internals*/
/*@{*/

/*! The operator precedence table*/
struct co_precedent_ops const * ops_tab;

/* Helpers ****************************************************************/

/*! Execute a binary operation on integer expressions.
	\param	lhs	An \e eval_result_t containing the left-hand integer operand.
	\param	rhs An \e eval_result_t containing the right-hand integer operand.
	\param	op	The binary operation to be applied to \e lhs and \e rhs.
	\return	An \e eval_result_t representing the result of evaluating
		\e op on \e lhs and \e rhs.

	It is not assumed that either of the expressions \e lhs or \e rhs has
	been resolved to an integer value. If either of them is unresolved then
	the returned result represents "the unresolved expression".
*/
static eval_result_t
integer_binary_op(	eval_result_t *lhs,
                    eval_result_t *rhs,
                    integer_bin_op_t op)
{
    eval_result_t result = EVAL_RESULT_INITOR;
    if ((IS_CONST(*lhs) || RESOLVED(*lhs)) &&
            (IS_CONST(*rhs) || RESOLVED(*rhs))) {
        int_spec_t int_spec = op(&lhs->value,&rhs->value);
        if (int_spec.type != INT_UNDEF) {
            eval_result_set_value(&result,&int_spec);
        }
    }
    return result;
}

/*! Less-than operator*/
static eval_result_t
op_lt(eval_result_t * lhs, eval_result_t * rhs)
{
    return integer_binary_op(lhs,rhs,lt);
}

/*! Greater-than operator*/
static eval_result_t
op_gt(eval_result_t *lhs, eval_result_t *rhs)
{
    return integer_binary_op(lhs,rhs,gt);
}

/*! Less-than-or-equal operator*/
static eval_result_t
op_le(eval_result_t *lhs, eval_result_t *rhs)
{
    return integer_binary_op(lhs,rhs,le);
}

/*! Greater-than-or-equal operator*/
static eval_result_t
op_ge(eval_result_t *lhs, eval_result_t *rhs)
{
    return integer_binary_op(lhs,rhs,ge);
}

/*! Equality operator*/
static eval_result_t
op_eq(eval_result_t *lhs, eval_result_t *rhs)
{
    return integer_binary_op(lhs,rhs,eq);
}

/*! Inequality operator*/
static eval_result_t
op_ne(eval_result_t *lhs, eval_result_t *rhs)
{
    return integer_binary_op(lhs,rhs,ne);
}

/*! Inclusive-or operator*/
static eval_result_t
op_or(eval_result_t *lhs, eval_result_t *rhs)
{
    eval_result_t result = EVAL_RESULT_INITOR;
    if (IS_TRUE(*lhs) || IS_TRUE(*rhs)) {
        eval_result_set_value(&result,&int_spec_true);
    } else if (IS_FALSE(*lhs) && IS_FALSE(*rhs)) {
        eval_result_set_value(&result,&int_spec_false);
    }
    if (FLAGS_IN(*lhs,EVAL_RESOLVED) == FLAGS_IN(*rhs,EVAL_RESOLVED)) {
        SET_FLAGS_IF(KEEP_CONST(*lhs) | KEEP_CONST(*rhs),
                     result,EVAL_KEEP_CONST);
    } else if (IS_TRUE(*lhs)) {
        SET_FLAGS_IF(KEEP_CONST(*lhs),result,EVAL_KEEP_CONST);
        CLEAR_FLAGS_IF(!KEEP_CONST(*rhs),*rhs,EVAL_KEEP);
    } else if (IS_TRUE(*rhs)) {
        SET_FLAGS_IF(KEEP_CONST(*rhs),result,EVAL_KEEP_CONST);
        CLEAR_FLAGS_IF(!KEEP_CONST(*lhs),*lhs,EVAL_KEEP);
    } else if (IS_FALSE(*lhs)) {
        SET_FLAGS_IF(KEEP_CONST(*rhs),result,EVAL_KEEP_CONST);
    } else if (IS_FALSE(*rhs)) {
        SET_FLAGS_IF(KEEP_CONST(*lhs),result,EVAL_KEEP_CONST);
    }
    return result;
}

/*! And operator*/
static eval_result_t
op_and(eval_result_t *lhs, eval_result_t *rhs)
{
    eval_result_t result = EVAL_RESULT_INITOR;
    if (IS_TRUE(*rhs) && IS_FALSE(*lhs)) {
        eval_result_set_value(&result,&int_spec_false);
        SET_FLAGS_IF(KEEP_CONST(*lhs),result,EVAL_KEEP_CONST);
    } else if (IS_TRUE(*lhs) && IS_FALSE(*rhs)) {
        eval_result_set_value(&result,&int_spec_false);
        SET_FLAGS_IF(KEEP_CONST(*rhs),result,EVAL_KEEP_CONST);
    } else if (UNRESOLVED(*lhs)) {
        if (IS_FALSE(*rhs)) {
            eval_result_set_value(&result,&int_spec_false);
            SET_FLAGS_IF(KEEP_CONST(*rhs),result,EVAL_KEEP_CONST);
        } else {
            SET_FLAGS_IF(KEEP_CONST(*lhs),result,EVAL_KEEP_CONST);
        }
    } else if (UNRESOLVED(*rhs)) {
        if (IS_FALSE(*lhs)) {
            eval_result_set_value(&result,&int_spec_false);
            SET_FLAGS_IF(KEEP_CONST(*lhs),result,EVAL_KEEP_CONST);
        } else {
            SET_FLAGS_IF(KEEP_CONST(*rhs),result,EVAL_KEEP_CONST);
        }
    } else {
        int_spec_t int_spec = boolean_and(&lhs->value,&rhs->value);
        eval_result_set_value(&result,&int_spec);
        SET_FLAGS_IF(KEEP_CONST(*rhs) | KEEP_CONST(*lhs),
                     result,EVAL_KEEP_CONST);
    }
    return result;
}

/*! Bitwise AND operator*/
static eval_result_t
op_bit_and(eval_result_t *lhs, eval_result_t *rhs)
{
    return integer_binary_op(lhs,rhs,bit_and);
}

/*! Bitwise OR operator*/
static eval_result_t
op_bit_or(eval_result_t *lhs, eval_result_t *rhs)
{
    return integer_binary_op(lhs,rhs,bit_or);
}

/*! Bitwise XOR operator*/
static eval_result_t
op_bit_xor(eval_result_t *lhs, eval_result_t *rhs)
{
    return integer_binary_op(lhs,rhs,bit_xor);
}

/*! Left-shift operator*/
static eval_result_t
op_lshift(eval_result_t *lhs, eval_result_t *rhs)
{
    return integer_binary_op(lhs,rhs,lshift);
}

/*! Right-shift operator*/
static eval_result_t
op_rshift(eval_result_t *lhs, eval_result_t *rhs)
{
    return integer_binary_op(lhs,rhs,rshift);
}

/*! Addition operator*/
static eval_result_t
op_add(eval_result_t *lhs, eval_result_t *rhs)
{
    return integer_binary_op(lhs,rhs,add);
}

/*! Substraction operator*/
static eval_result_t
op_subtract(eval_result_t *lhs, eval_result_t *rhs)
{
    return integer_binary_op(lhs,rhs,subtract);
}

/*! Multiplication operator*/
static eval_result_t
op_mult(eval_result_t *lhs, eval_result_t *rhs)
{
    return integer_binary_op(lhs,rhs,multiply);
}

/*! Division operator*/
static eval_result_t
op_divide(eval_result_t *lhs, eval_result_t *rhs)
{
    eval_result_t result = EVAL_RESULT_INITOR;
    if ((IS_CONST(*lhs) || RESOLVED(*lhs)) &&
            (IS_CONST(*rhs) || RESOLVED(*rhs))) {
        if (!rhs->value.val.ull) {
            report(GRIPE_ZERO_DIVIDE,NULL,
                   "Divide by zero");
        } else {
            int_spec_t int_spec = divide(&lhs->value,&rhs->value);
            eval_result_set_value(&result,&int_spec);
        }
    }
    return result;
}

/*! Modulus operator*/
static eval_result_t
op_mod(eval_result_t *lhs, eval_result_t *rhs)
{
    eval_result_t result = EVAL_RESULT_INITOR;
    if ((IS_CONST(*lhs) || RESOLVED(*lhs)) &&
            (IS_CONST(*rhs) || RESOLVED(*rhs))) {
        if (!rhs->value.val.ull) {
            report(GRIPE_ZERO_DIVIDE,NULL,
                   "Divide by zero");
        } else {
            int_spec_t int_spec = modulus(&lhs->value,&rhs->value);
            eval_result_set_value(&result,&int_spec);
        }
    }
    return result;
}

/*! Type of evaluation functions.

	\param		ops 	A pointer to an element of the precedence
						table which lists the	operators at the
						current level of precedence.

	\param		cpp On entry, a pointer to a \c char* that points to the
				expression to be evaluated. On return, receives
				the address terminating the evaluated expression.

	\return An \em eval_result_t representing the result of
				evaluation.
*/
typedef eval_result_t (evaluator_t)(
    struct co_precedent_ops const * ops,
    char const **);

/*! Evaluator for innermost subexpressions, i.e.
 - \em !expr
 - \em (expr)
 - \em (symbol)
 - \em symbol
 - \em number

 \param	ops		Pointer to the \em struct ops to be used for
					the evaluation.
 \param cpp		On entry, a pointer to the start of the text
					to be evaluated. Receives the address reached
					by evaluation.
 \return	An \em eval_result_t representing
			the result of evaluation.
*/
static eval_result_t
eval_unary(struct co_precedent_ops const * ops, char const **cpp);


/*! Table-driven evaluation of binary operators.
  The evaluator does shortcircuit evaluation for && and ||.
  It also simplifies && and || subexpressions that
  can be partially solved.
  If we are keeping constants then constants are treated like
  unresolved symbols, and the rules are:
	CONSTANT1 && CONSTANT2 := CONSTANT1
	CONSTANT && TRUE : = CONSTANT
	CONSTANT && FALSE := FALSE
	CONSTANT1 || CONSTANT2 := CONSTANT1
	CONSTANT || TRUE := TRUE
	CONSTANT || FALSE := CONSTANT

 \param	ops		Pointer to the \em struct ops to be used for
					the evaluation.
 \param cpp		On entry, a pointer to the start of the text
					to be evaluated. Receives the address reached
					by evaluation.
 \return		An \em eval_result_t that representing the result of
				evaluation.
*/
static eval_result_t
eval_table(struct co_precedent_ops const *ops, char const **cpp);


/*! Structure representing an operator
	(Would rather call it "operator" but GCC thinks
	that is a reserved word even in C)*/

struct operation {
    /*! Points to the preprocessor token that denotes the operator*/
    const char *str;
    /*! Pointer to function to implement the operator*/
    eval_result_t (*fn)(eval_result_t *, eval_result_t *);
};

/*! Structure of the operator precedence table.
 *
    Expressions involving binary operators are evaluated
    in a table-driven way by \c eval_table(). When it evaluates a subexpression
    it calls the inner function with its first argument pointing to the next
    element of the table.
 *
    Innermost expressions have special non-table-driven handling.
*/
struct co_precedent_ops {
    /*! Pointer to function for evaluating operands of operator*/
    evaluator_t *inner;
    /*! Array of operators of same precedence associated with
    	the evaluator \c inner*/
    struct operation op[5];
};


/*! The operator precedence table*/
const struct co_precedent_ops eval_ops[] = {

    {
        eval_table,
        {
            { BOOLEAN_OR, op_or }
        }
    },
    {
        eval_table,
        {
            { BOOLEAN_AND, op_and }
        }
    },
    {
        eval_table,
        {
            { BIT_OR, op_bit_or },
            { BIT_XOR, op_bit_xor },
            { BIT_AND, op_bit_and }
        }
    },
    {
        eval_table,
        {
            { EQ, op_eq },
            { NEQ, op_ne }
        }
    },
    {
        eval_table,
        {
            { LE, op_le },
            { GE, op_ge },
            { LESS, op_lt },
            { GREATER, op_gt }
        }
    },
    {
        eval_table,
        {
            { LSHIFT, op_lshift },
            { RSHIFT, op_rshift },
        }
    },
    {
        eval_table,
        {
            { ADD, op_add },
            { SUBTRACT, op_subtract },
        }
    },
    {
        eval_unary,
        {
            { MULTIPLY, op_mult },
            { DIV, op_divide },
            { MOD, op_mod },
        }
    },
};

struct co_precedent_ops const * ops_tab = eval_ops;

/*!	Skip right-hand operand of \c && in an hash-if
 *  directive when we can short-circuit the evaluation.
 *
    \param cp	Line-buffer position at which the operand
    			begins.
 *
    When an \c && can be short-circuited, the evaluator
    can skip from the start of the righthand operand until:
    - the end of the directive, or
    - an unbalanced ')', which must close the scope of the \c &&, or
    - a \c || operator, which has higher precedence.
 *
    \return The address just past the skipped operand.
*/
static char const *
short_circuit_and(char const *cp)
{
    int balance = 0;
    for (	; *(cp = chew_on(cp)); ++cp) {
        if (*cp == '(') {
            ++balance;
        } else if (*cp == ')') {
            --balance;
        }
        if (balance < 0) {
            break;
        } else if (balance == 0) {
            /* Check for another binary truth functor on same
             * level to respect precedence.
            */
            if (cp[0] == cp[1] && cp[0] == '|') {
                break;
            }
        }
    }
    return cp;
}

/*!	Skip the right-hand operand of \c || in an hash-if
 *  directive when we can short-circuit the evaluation.
 *
    \param cp	Line-buffer position at which the operand
    			begins.
 *
    When an \c || can be short-circuited, the evaluator can
    skip from the start of the righthand operand until the end
    of the directive, or an unbalanced ')', which must close
    the scope of the \c ||. None of our operators has higher
    precedence.
 *
    \return The address just past the skipped operand.
*/
static char const *
short_circuit_or(char const *cp)
{
    int balance = 0;
    for (	; *(cp = chew_on(cp)); ++cp) {
        if (*cp == '(') {
            ++balance;
        } else if (*cp == ')') {
            --balance;
        }
        if (balance < 0) {
            break;
        }
    }
    return cp;
}

/*! Evaluate the expression of a hash-if or hash-elif directive.
    If we can resolve the expression we return \c LT_TRUE or
    \c LT_FALSE accordingly, otherwise \c LT_IF.
*/
static int
eval_if(char const **cpp);

/*!	Evaluate the text of a hash-define directive to determine if it is
	consistent with the \c --define and \c --undefine assumptions.

	\param cpp  Pointer to the address of the directive text. This
				pointer is updated at exit with the address just past the
				text consumed.

	\return		A line-type, as follows:

	- Case 1: Directive is hash-define \c sym

		-	If neither \c --define \c sym nor \c --undefine \c sym,
			return \c LT_CONSISTENT_DEFINE_KEEP
		-	If \c --define \c sym, return \c LT_CONSISTENT_DEFINE_DROP
		-	If <tt>--define sym=val</tt>, return \c LT_DIFFERING_DEFINE
		-	If \c --undefine \c sym, return \c LT_CONTRADICTORY_DEFINE

	- Case 2: Directive is hash-define sym str1 [str2...]

		-	If neither \c --define \c sym nor \c --undefine \c sym,
			return \c LT_CONSISTENT_DEFINE_KEEP
		-	If \c --define \c sym, return \c LT_DIFFERING_DEFINE
		-	If <tt>--define sym=str1</tt>, return \c LT_CONSISTENT_DEFINE_KEEP
		-	If <tt>--define sym=val</tt>, where <tt>val != str1</tt>,
				return \c LT_DIFFERING_DEFINE
		-	If \c --undefine \c sym return \c LT_CONTRADICTORY_DEFINE

	-	Case 3: Directive is hash-define sym(arg1[,arg2...]) str1 [str2...]

		-	If neither \c --define \c sym nor \c --undefine \c sym,
			return \c LT_CONSISTENT_DEFINE_KEEP
		-	If \c --undefine \c sym, return \c LT_CONTRADICTORY_DEFINE
		-	If <tt>--define sym=val</tt> return \c LT_DIFFERING_DEFINE
		-	If \c --define \c sym, return \c LT_DIFFERING_DEFINE
*/
static int
eval_hash_define(char const **cpp)
{
    canonical_string_const_h old_def = NULL;
    canonical_string_h symname = NULL;
    char const * cp = *cpp;
    char const *directive;
    int retval = LT_PLAIN;
    symbol_h sym;
    debug(DBG_13,line_len(*cpp),*cpp);
    cp = directive = chew_on(cp);
    if (eligible_line(CMD_DEFS,HASH_DEFINE)) {
        canonical_string_h def_text = canonical_string_new(&directive,0);
        directive_tally(def_text,HASH_DEFINE);
    }
    symname = canonical_identifier(&cp);
    sym = configured_symbol_match(symname);
    do {
        if (sym) {
            if (eligible_line(CMD_SYMBOLS,HASH_DEFINE)) {
                configured_symbol_tally(sym);
            }
            if (GET_PUBLIC(line_select,dropping_line)) {
                break;
            }
            old_def = symbol_definition(sym);
        } else {
            if (eligible_line(CMD_SYMBOLS,HASH_DEFINE)) {
                unconfigured_symbol_tally(canonical_string_copy(symname));
            }
            if (GET_PUBLIC(line_select,dropping_line)) {
                break;
            }
            if (GET_PUBLIC(args,implicit)) {
                retval = LT_CONTRADICTORY_DEFINE;
            } else {
                retval = LT_CONSISTENT_DEFINE_KEEP;	/* symbol not -Ded or -Ued*/
            }
            break;
        }
        if (old_def == NULL) {
            retval = LT_CONTRADICTORY_DEFINE;	/* symbol is -Ued*/
            break;
        }
        cp = chew_on(cp);
        if (*cp != '\0') {
            canonical_string_h new_def = canonical_string_new(&cp,0);
            retval = canonical_string_equal(new_def,old_def) ?
                     LT_CONSISTENT_DEFINE_DROP :	LT_DIFFERING_DEFINE;
            canonical_string_dispose(new_def);
            break;
        }
        /* #define sym*/
        if (canonical_string_length(old_def) == 0) {
            if (sym == GET_PUBLIC(categorical,last_contradictory_undef)) {
                forget_contradiction();
            }
            retval = LT_CONSISTENT_DEFINE_DROP;
        } else {
            retval = LT_DIFFERING_DEFINE;
        }
    } while(false);
    while(*cp) { /* Chew to a line end */
        cp = chew_string(cp);
        cp = chew_on(cp);
    }
    canonical_string_dispose(symname);
    debug(DBG_14,line_len(*cpp),*cpp,retval);
    *cpp = cp;
    return retval;
}

/*!	Evaluate the text of an hash-undef to determine if it is consistent
	with the \c --define and \c --undefine options.

	\param cpp  Pointer to the address of the directive text. This
				pointer is updated at exit with the address just past the
				text consumed.

	\return		A line-type, as follows:

	- 	If neither <tt>--define sym</tt> nor <tt>--undefine sym</tt>,
		return \c LT_CONSISTENT_UNDEF_KEEP
	-	If \c --undefine \c sym, return \c LT_CONSISTENT_UNDEF_DROP
	-	Else return \c LT_CONTRADICTORY_UNDEF
*/
static int
eval_hash_undef(char const **cpp)
{
    canonical_string_const_h def = NULL;
    canonical_string_h symname = NULL;
    char const *cp = *cpp;
    int retval = LT_PLAIN;
    symbol_h sym;
    debug(DBG_15,line_len(*cpp),*cpp);
    cp = chew_on(cp);
    symname = canonical_identifier(&cp);
    if (eligible_line(CMD_DEFS,HASH_UNDEF)) {
        directive_tally(canonical_string_copy(symname),HASH_UNDEF);
    }
    sym = configured_symbol_match(symname);
    do {
        if (sym ) {	/* Symbol is -Ded or -Ued*/
            if (eligible_line(CMD_SYMBOLS,HASH_UNDEF)) {
                configured_symbol_tally(sym);
            }
            if (GET_PUBLIC(line_select,dropping_line)) {
                break;
            }
            def = symbol_definition(sym);
        } else {
            if (eligible_line(CMD_SYMBOLS,HASH_UNDEF)) {
                unconfigured_symbol_tally(canonical_string_copy(symname));
            }
            if (GET_PUBLIC(line_select,dropping_line)) {
                break;
            }
            retval = LT_CONSISTENT_UNDEF_KEEP;	/* symbol not -Ded or -Ued*/
            break;
        }
        if (def == NULL) {
            retval = LT_CONSISTENT_UNDEF_DROP;	/* symbol is -Ued*/
        } else {	/* Symbol is -Ded*/
            /* This #undef FOO contradicts -DFOO. We will complain
            about this unless #define FOO is the next thing we see after
            comments and whitespace, so we stash the complaint for the
            the time being - unless we've already got a pending
            contradiction for the same symbol*/
            if (GET_PUBLIC(categorical,last_contradictory_undef) != sym) {
                flush_contradiction();
                SET_PUBLIC(categorical,last_contradictory_undef) = sym;
                save_contradiction("\"%.*s\" contradicts -D symbol");
            }
            retval = LT_CONTRADICTORY_UNDEF;
        }
    } while(false);
    debug(DBG_16,line_len(*cpp),*cpp,retval);
    canonical_string_dispose(symname);
    *cpp = cp = chew_on(cp);
    return retval;
}


/*!	Evaluate the argument of a hash-include directive.
	\param	cp Input text pointer at which to start parsing.
	\return Pointer to the input character at which parsing stops.

	The function is called only when the \c includes command is in
	effect.

	If the argument of the hash-include directive is a malformed header
	name the function will report a warning diagnostic on stderr.
	Likewise if the argument is a macro call, since macro expension
	is not supported, or if the argument is preprocessor symbol that is:-

	- unknown
	- undef-ed
	- evaluates to an empty string.
	- evaluates to a string that is a malformed header name.

	The function will list the hash-include directive on stdout in
	all circumstances if \c --once is not in effect. If \c --once
	is effect then the directive will be listed if it has not already
	been seen.
*/
static char const *
eval_hash_include(char const *cp)
{
    canonical_string_h resolved_val = NULL;
    char const *diagnostic = NULL;
    char const *arg_end;
    size_t arglen;
    SET_PUBLIC(chew,stop_at_quote) = true;
    cp = chew_on(cp);
    if (!*cp) {
        report(GRIPE_INVALID_INCLUDE,NULL,
               "#include has no argument");
        return cp;
    }
    arg_end = hash_include_parse(cp,&diagnostic,&resolved_val);
    arglen = arg_end - cp;
    if (diagnostic) { /* Something problematic */
        if (resolved_val) { /* Argument is a configured symbol */
            report(GRIPE_INVALID_INCLUDE,NULL,
                   "\"#include %.*s\": defined symbol resolved as (%s): %s",
                   arglen,cp,canonical_string_text(resolved_val),diagnostic);
        } else {
            report(GRIPE_INVALID_INCLUDE,NULL,
                   "\"#include %.*s\": %s",arglen,cp,diagnostic);
        }

    }
    if (eligible_line(CMD_INCLUDES,HASH_INCLUDE)) {
        /* Do the reporting for the \c includes command */
        hash_include_tally(cp,arglen,resolved_val);
    }
    return arg_end;
}

/*! Evaluate a directive that takes a free-form argumnent
	(hash-error, hash-pragma)
	\param	cp	Pointer to the argument of the hash-error directive.
	\param	directive_type Type of the directive being evaluated.
	\return	The terminal address of the hash-error directive.
*/
static char const *
eval_free_form_directive(char const *cp, directive_type_t directive_type)
{
    canonical_string_h cs = canonical_string_new(&cp,0);
    directive_tally(cs,directive_type);
    return cp;
}


static eval_result_t
eval_table(struct co_precedent_ops const *ops, char const **cpp)
{
    char const *start_cut = (debug(DBG_8, ops - eval_ops),*cpp);
    /* Evaluate the lhs...*/
    eval_result_t lhs_result = ops->inner(ops+1,cpp);
    /* Assume lhs is all we've got...*/
    eval_result_t result = lhs_result;
    char const * cp = *cpp;
    if (LEGAL(result)) {
        /* Assume we will delete parentheses...*/
        SET_DEL_PAREN(result);
        for (; *(cp = chew_on(cp)) && *cp != ')'; ) {
            eval_result_t rhs_result = EVAL_RESULT_INITOR;
            char const *start_lhs_cut = start_cut;
            char const *end_lhs_cut = cp;
            const struct operation *op;

            /* Now look for binary op at current precedence...*/
            for (op = ops->op; op->str != NULL && !match_op(&cp,op->str); op++) {}
            if (op->str == NULL) {
                /* No binary op at current precedence, skip rhs. */
                if (ops == eval_ops) {
                    /* No binary op at all. Illegal */
                    SET_ILLEGAL(result);
                }
                break;
            }
            /* Got bin op - we will not delete parentheses...*/
            CLEAR_FLAGS(result,EVAL_DEL_PAREN);
            if (RESOLVED(lhs_result) && !KEEP_CONST(lhs_result)) {
                /* ...Possibly can short-circuit*/
                if (IS_TRUE(lhs_result) && op->fn == op_or) {
                    /*	Can shortcircuit on TRUE || <RHS>...*/
                    cp = short_circuit_or(cp);
                    /* result == lhs_result. OK*/
                    break;
                } else if (IS_FALSE(lhs_result) && op->fn == op_and) {
                    /*	Can shortcircuit on FALSE && <RHS>...*/
                    cp = short_circuit_and(cp);
                    /* result == lhs_result. OK*/
                    break;
                } else if ((IS_TRUE(lhs_result) && op->fn == op_and) ||
                           (IS_FALSE(lhs_result) && op->fn == op_or)) {
                    /* Can delete TRUE &&, or FALSE ||...*/
                    cut_text(LINE_OFF(start_lhs_cut),LINE_OFF(cp));
                    /* Cutting lhs can make superflous parentheses*/
                    SET_DEL_PAREN(result);
                }
            }
            start_cut = end_lhs_cut;
            end_lhs_cut = cp;
            debug(DBG_9, ops - eval_ops, op->str);
            /* Evaluate rhs...*/
            rhs_result = ops->inner(ops,&cp);
            result = op->fn(&lhs_result,&rhs_result);
            if ((op->fn == op_or || op->fn == op_and)) {
                if (!KEEP(lhs_result) && KEEP(rhs_result)) {
                    cut_text(LINE_OFF(start_lhs_cut),LINE_OFF(end_lhs_cut));
                    SET_FLAGS(result,EVAL_DEL_PAREN | EVAL_KEEP);
                } else if (!KEEP(rhs_result) && KEEP(lhs_result)) {
                    cut_text(LINE_OFF(start_cut),LINE_OFF(cp));
                    SET_FLAGS(result,EVAL_DEL_PAREN | EVAL_KEEP);
                }
            }
        }
    }
    if (RESOLVED(result)) {
        debug(DBG_10, ops - eval_ops, result.value);
    } else {
        SET_KEEP(result);
    }
    *cpp = cp;
    return result;
}

static eval_result_t
eval_unary(struct co_precedent_ops const *ops, char const **cpp)
{
    char const *ep;
    eval_result_t result = EVAL_RESULT_INITOR;
    char const *cp;
    SET_PUBLIC(chew,stop_at_quote) = true;
    cp = chew_on(*cpp);
    SET_PUBLIC(chew,stop_at_quote) = false;
    canonical_string_h symname = NULL;
    do {

        if (*cp == '!') {
            debug(DBG_1,ops - eval_ops);
            cp = chew_continuation(++cp);
            result = eval_unary(ops,&cp);
            if (UNRESOLVED(result)) {
                break;
            }
            result.value = negate(&result.value);
            FLIP_FLAGS(result,EVAL_RESOLVED);
        } else if (*cp == '~') {
            int_spec_t val;
            debug(DBG_22,ops - eval_ops);
            cp = chew_continuation(++cp);
            result = eval_unary(ops,&cp);
            if (UNRESOLVED(result)) {
                break;
            }
            val = complement(&result.value);
            eval_result_set_value(&result,&val);
        } else if (*cp == '(') {
            char const *start = cp;
            cp = chew_continuation(++cp);
            debug(DBG_2,ops - eval_ops);
            result = eval_table(eval_ops,&cp);
            cp = chew_on(cp);
            if (*cp != ')') {
                if (GET_STATE(evaluator,parsing_sym_def_depth) == 0) {
                    /* Missing ')'*/
                    bail(GRIPE_UNBALANCED_PAREN,
                         "Missing \")\" in \"%.*s\"",
                         line_len(GET_PUBLIC(io,pristine_line)),
                         GET_PUBLIC(io,pristine_line));
                } else {
                    SET_ILLEGAL(result);
                    break;
                }
            } else if (DEL_PAREN(result)) {
                delete_paren(LINE_OFF(start),LINE_OFF(cp));
            }
            cp = chew_continuation(++cp);
        } else if (*cp == '+') {
            debug(DBG_20,ops - eval_ops);
            cp = chew_continuation(++cp);
            result = eval_unary(ops,&cp);
            break;
        } else if (*cp == '-') {
            debug(DBG_21,ops - eval_ops);
            cp = chew_continuation(++cp);
            result = eval_unary(ops,&cp);
            result.value = minus(&result.value);
        } else if (isdigit((unsigned char)*cp)) {
            int_spec_t val;
            debug(DBG_3,ops - eval_ops);
            val = eval_numeral(cp,&ep);
            if (val.type != INT_TOO_BIG) {
                result.value = val;
                if (GET_STATE(evaluator,parsing_sym_def_depth) == 0) {
                    SET_CONST(result);
                    if (GET_PUBLIC(args,eval_consts)) {
                        SET_FLAGS(result,result.value.val.ull ? EVAL_TRUE : EVAL_FALSE);
                    }
                } else {
                    SET_FLAGS(result,result.value.val.ull ? EVAL_TRUE : EVAL_FALSE);
                }
            }
            cp = (char *)ep;
        } else if (match_keyword(&cp,"defined")) {
            symbol_h sym;
            bool paren;
            cp = chew_on(cp);
            debug(DBG_4, ops - eval_ops);
            paren = *cp == '(';
            if (paren) {
                cp = chew_continuation(++cp);
            }
            cp = chew_on(cp);
            symname = canonical_identifier(&cp);
            sym = configured_symbol_match(symname);
            cp = chew_on(cp);
            if (paren) {
                if (*cp == ')') {
                    cp = chew_continuation(++cp);
                }
                /* Missing ')'*/
                else if (GET_STATE(evaluator,parsing_sym_def_depth) == 0) {
                    bail(GRIPE_UNBALANCED_PAREN,NULL,
                         "Missing \")\" in \"%.*s\"",
                         line_len(GET_PUBLIC(io,pristine_line)),
                         GET_PUBLIC(io,pristine_line));
                } else {
                    SET_ILLEGAL(result);
                    break;
                }
            }
            if (!sym) {
                if (eligible_line(CMD_SYMBOLS,HASH_IF)) {
                    unconfigured_symbol_tally(canonical_string_copy(symname));
                }
		        if (GET_PUBLIC(args,implicit)) {
                    eval_result_set_value(&result,&int_spec_false);
				}
                break;
            } else if (eligible_line(CMD_SYMBOLS,HASH_IF)) {
                configured_symbol_tally(sym);
            }
            eval_result_set_value(&result,(symbol_definition(sym) != NULL) ? &int_spec_true : &int_spec_false );
        } else if (*cp == '\'' || cp[0] == 'L' ) { /* Character constant */
            int_spec_t val = eval_character_constant(&cp);
            if (val.type != INT_INSOLUBLE) { /* Got a character constant */
                eval_result_set_value(&result,&val);
            }
        } else if (is_symbol_start_char(*cp)) {
            symbol_h sym;
            char const *arg_start = cp;
            char const *arg_end;
            debug(DBG_5,ops - eval_ops);
            symname = canonical_identifier(&cp);
            arg_end = chew_macro_call(arg_start);
            if (arg_end > cp) {
                report(GRIPE_MACROS_UNSUPPORTED,NULL,
                       "Sorry. Macro-expansion is not supported: \"%.*s\"",arg_end - arg_start,arg_start);
            }
            cp = arg_end;
            sym = configured_symbol_match(symname);
            if (!sym) {
                if (eligible_line(CMD_SYMBOLS,HASH_IF) &&
                        GET_STATE(evaluator,parsing_sym_def_depth) == 0) {
                    unconfigured_symbol_tally(canonical_string_copy(symname));
                }
		        if (GET_PUBLIC(args,implicit)) {
                    eval_result_set_value(&result,&int_spec_false);
				}
                /* Unknown symbol*/
                break;
            } else if (eligible_line(CMD_SYMBOLS,HASH_IF) &&
                       GET_STATE(evaluator,parsing_sym_def_depth) == 0) {
                configured_symbol_tally(sym);
            }
            symbol_resolve(sym);
            if (!symbol_def_len(sym)) {
                if (GET_STATE(evaluator,parsing_sym_def_depth) == 0) {
                    report(GRIPE_EMPTY_SYMBOL,NULL,
                           "Empty symbol \"%s\" in expression",symbol_name(sym));
                }
            }
            result = *sym;
        } else {
            SET_ILLEGAL(result);
            debug(DBG_6,ops - eval_ops);
            break;
        }
    } while(false);
    *cpp = cp;
    if (RESOLVED(result)) {
        debug(DBG_7,ops - eval_ops,result.value);
    } else {
        SET_KEEP(result);
    }
    canonical_string_dispose(symname);
    return result;
}


static int
eval_if(char const **cpp)
{
    eval_result_t result;
    debug(DBG_11, *cpp);
    result = eval_table(eval_ops,cpp);
    debug(DBG_12,result.value);

    if (KEEP_CONST(result)) {
        return LT_IF;
    }
    if (IS_TRUE(result)) {
        return LT_TRUE;
    }
    if (IS_FALSE(result)) {
        return LT_FALSE;;
    }
    return LT_IF;
}

/*@}*/

/* API **************************************************************/

line_type_t
eval_line(void)
{
    char const *cp;
    symbol_h sym;
    size_t kwlen;
    int retval;
    comment_state_t wascomment;

    if (!get_line()) {
        flush_contradiction();
        transition(LT_EOF);
        return LT_EOF;
    }

    SET_PUBLIC(line_edit,simplification_state) = UNSIMPLIFIED;
    /* Assume no simplification possible*/
    retval = LT_PLAIN;
    wascomment = GET_PUBLIC(chew,comment_state);
    cp = chew_on(GET_PUBLIC(io,line_start));
    if (GET_PUBLIC(chew,line_state) == LS_NEUTER) {
        if (*cp == '#') {
            SET_PUBLIC(chew,line_state) = LS_DIRECTIVE;
            cp = chew_on(cp + 1);
        } else if (*cp != '\0') {
            SET_PUBLIC(chew,line_state) = LS_CODE;
            flush_contradiction();
        }
    }
    if (GET_PUBLIC(chew,comment_state) == NO_COMMENT &&
            GET_PUBLIC(chew,line_state) == LS_DIRECTIVE) {
        size_t kwoff = read_offset(cp);
        char *kwpos;
        directive_type_t directive;
        SET_PUBLIC(chew,stop_at_quote) = false;
        SET_PUBLIC(chew,must_balance_quotes) = true;
        SET_STATE(evaluator,parsing_sym_def_depth) = 0;
        kwpos = read_pos(kwoff);
        SET_PUBLIC(line_edit,keyword) = kwpos;
        directive = get_directive_type(&cp);
        kwlen = cp - kwpos;
        SET_PUBLIC(line_select,dropping_line) = dropping_line();
        switch(directive) {
        case HASH_IF:
            SET_STATE(evaluator,ifpos) = cp;
            retval = eval_if(&cp);
            break;
        case HASH_IFDEF:
        case HASH_IFNDEF: {
            canonical_string_h symname = NULL;
            cp = chew_on(cp);
            symname = canonical_identifier(&cp);
            sym = configured_symbol_match(symname);
            retval = LT_IF;
            if (sym) {
                if (eligible_line(CMD_SYMBOLS,directive)) {
                    configured_symbol_tally(sym);
                }
                if (SYM_DEFINED(sym)) {
                    /* symbol is -Ded*/
                    retval = directive == HASH_IFDEF ? LT_TRUE : LT_FALSE;
                } else {
                    /* symbol is -Ued*/
                    retval = directive == HASH_IFNDEF ? LT_TRUE : LT_FALSE;
                }
            } else {
                if (eligible_line(CMD_SYMBOLS,directive)) {
                    unconfigured_symbol_tally(canonical_string_copy(symname));
                }
                if (GET_PUBLIC(args,implicit)) {
                    retval = directive == HASH_IFDEF ? LT_FALSE : LT_TRUE;
                }
            }
            canonical_string_dispose(symname);
        }
        break;
        case HASH_ELSE:
            retval = LT_ELSE;
            break;
        case HASH_ELIF:
            retval = eval_if(&cp) - LT_IF + LT_ELIF;
            break;
        case HASH_ENDIF:
            retval = LT_ENDIF;
            break;
        case HASH_DEFINE:
            retval = eval_hash_define(&cp);
            break;
        case HASH_UNDEF:
            retval = eval_hash_undef(&cp);
            break;
        case HASH_INCLUDE:
            cp = eval_hash_include(cp);
            retval = GET_PUBLIC(line_select,dropping_line) ? LT_INCLUDE_DROP : LT_INCLUDE_KEEP;
            break;
        case HASH_PRAGMA:
            SET_PUBLIC(chew,must_balance_quotes) = false;
            if (eligible_line(CMD_PRAGMAS,HASH_PRAGMA)) {
                cp = eval_free_form_directive(cp,HASH_PRAGMA);
            }
            break;
        case HASH_ERROR:
            SET_PUBLIC(chew,must_balance_quotes) = false;
            if (eligible_line(CMD_ERRORS,HASH_ERROR)) {
                cp = eval_free_form_directive(cp,HASH_ERROR);
            }
            if (!GET_PUBLIC(line_select,dropping_line)) {
                SET_PUBLIC(chew,last_comment_start_line) =
                    GET_PUBLIC(io,line_num);
                if (is_unconditional_line()) {
                    set_exit_flags(EVENT_SUMMARY_ERROR_OUTPUT,true);
                    if (was_unconditional_line()) {
                        report(GRIPE_UNCONDITIONAL_ERROR_INPUT,NULL,
                               "An unconditional #error directive was input");
                    } else {
                        report(GRIPE_UNCONDITIONAL_ERROR_INPUT,NULL,
                               "An unconditional #error directive becomes operative");
                    }
                }
            }
            break;
        default:
            report(GRIPE_UNKNOWN_DIRECTIVE,NULL,"Unknown directive #%.*s",
                   kwlen,kwpos);
            SET_PUBLIC(chew,line_state) = LS_CODE;
            retval = LT_PLAIN;
        };
        cp = chew_on(cp);
        if (*cp) {
            SET_PUBLIC(chew,line_state) = LS_CODE;
            if (retval != LT_PLAIN) {
                if (*cp == ')') {
                    bail(GRIPE_UNBALANCED_PAREN,
                         "Missing \"(\" in \"%.*s\"",
                         line_len(GET_PUBLIC(io,pristine_line)),
                         GET_PUBLIC(io,pristine_line));
                } else {
                    report(GRIPE_GARBAGE_AFTER_DIRECTIVE,NULL,
                           "Garbage following preprocessor directive in \"%.*s\"",
                           line_len(GET_PUBLIC(io,pristine_line)),
                           GET_PUBLIC(io,pristine_line));
                }
            }
        }
        if (retval != LT_CONTRADICTORY_UNDEF) {
            flush_contradiction();
        }
    }
    if (GET_PUBLIC(chew,line_state) == LS_CODE) {
        while (*cp != '\0') {
            cp = chew_on(cp + 1);
        }
    }
    debug(DBG_17);
    return (retval);
}

eval_result_t
eval_definition_text(char const **pdef)
{
    eval_result_t er;
    ++SET_STATE(evaluator,parsing_sym_def_depth);
    SET_PUBLIC(chew,in_source) = false;
    er = eval_table(eval_ops,pdef);
    SET_PUBLIC(chew,in_source) = true;
    /* Forget any unclosed quotation in symbol definition */
    SET_PUBLIC(chew,in_double_quote) = false;
    --SET_STATE(evaluator,parsing_sym_def_depth);
    return er;
}

/* EOF*/
