/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "integer_ops.h"
#include "report.h"

/*!\ingroup integer_ops_component integer_ops_interface integer_ops_internals
 *\file integer_ops.c
 * This file implements the Integer Operations component
*/

/*! \addtogroup integer_ops_internals*/
/*@{*/

/* Helpers */

/*! Return the sign of the value represented by an \em int_spec_t object
    \param  int_spec    The \e int_spec_t from which the sign is required.
    \return -1 if the value is of signed integral type and is negative, 1 if
        the value is of signed integral type and is non-negative; otherwise 0.
*/
static int
sign(int_spec_t const *int_spec)
{
    switch(int_spec->type) {
    case INT_INT:
        return int_spec->val.i  < 0 ? -1 : 1;
    case INT_UINT:
        return 1;
    case INT_LONG:
        return int_spec->val.l  < 0L ? -1 : 1;
    case INT_ULONG:
        return 1;
    case INT_LLONG:
        return int_spec->val.ll  < 0LL ? -1 : 1;
    case INT_ULLONG:
        return 1;
    default:
        return 0;
    }
}


/*! Apply sign extension to a type-converted \e int_spec_t
    \param  converted    Pointer to an \e int_spec_t having the
        correct converted type and the pre-sign-extended value.
    \param  unconverted Pointer to an \e int_spec_t having the uncoverted
            value and type.

    If \c unconverted does not represent a signed type with integer conversion rank > than that
    of \c converted and with a negative value then nothing is done to \c converted.
    Otherwise the uncoverted value from \c uncoverted is
    assigned to the field in \e converted corresponding to the converted type,
    thus accomplishing sign-extension if required.

*/
static void
sign_extend(int_spec_t * converted, int_spec_t const * unconverted)
{
    if (IS_SIGNED(unconverted->type) && sign(unconverted) == -1 &&
            RANK(converted->type) > RANK(unconverted->type)) {
        switch(converted->type) {
        case INT_LONG:
        case INT_ULONG:
            switch(unconverted->type) {
            case INT_INT:
                converted->val.ul = unconverted->val.i;
                break;
            default:
                assert(false);
            }
            break;
        case INT_LLONG:
        case INT_ULLONG:
            switch(unconverted->type) {
            case INT_INT:
                converted->val.ull = unconverted->val.i;
                break;
            case INT_LONG:
                converted->val.ull = unconverted->val.l;
                break;
            default:
                assert(false);
            }
            break;
        default:
            assert(false);
        }
    }
}

/*! Apply the Usual Arithmetic Conversions to a pair of \em int_spec_t objects
    to determine the result type of a binary integer operations with these
    objects as operands.
        See standard ISO/IEC 9899:1999 6.3.1.8

    \param l   Pointer to an \em int_spec_t representing the unconverted left-hand operand.
    \param r   Pointer to an \em int_spec_t representing the unconverted right-hand operand.
    \param cl  Pointer to an \em int_spec_t representing the converted left-hand operand on return.
    \param cr  Pointer to an \em int_spec_t representing the converted right-hand operand on return.

    Sign extension is performed on the converted operands if required.
*/

static void
do_usual_conversions(int_spec_t const * l, int_spec_t const *r, int_spec_t *cl, int_spec_t *cr)
{
    int_type lhs = l->type;
    int_type rhs = r->type;
    int_spec_t * post_sign_changed = NULL;
    int_spec_t const * pre_sign_changed = NULL;

    /* Initialise the converted operands with the unconverted operands */
    *cl = *l;
    *cr = *r;

    /* if both operands have the same type, then no further conversion is needed. */
    if (lhs == rhs) {
        return;
    }
    /* Otherwise, if both operands have signed integer types or both have unsigned
    integer types, the operand with the type of lesser integer conversion rank is
    converted to the type of the operand with greater rank. */
    if (IS_SIGNED(lhs) == IS_SIGNED(rhs)) {
        cl->type = cr->type = RANK(lhs) > RANK(rhs) ? lhs : rhs;
        return;

    }
    /* Otherwise, if the operand that has unsigned integer type has rank greater or
    equal to the rank of the type of the other operand, then the operand with
    signed integer type is converted to the type of the operand with unsigned
    integer type. */
    if (!IS_SIGNED(lhs) && RANK(lhs) >= RANK(rhs)) {
        post_sign_changed = cr;
        pre_sign_changed = r;
        cr->type = cl->type;
    }
    else if (!IS_SIGNED(rhs) && RANK(rhs) >= RANK(lhs)) {
        post_sign_changed = cl;
        pre_sign_changed = l;
        cl->type = cr->type;
    }
    /* Otherwise, if the type of the operand with signed integer type can represent
    all of the values of the type of the operand with unsigned integer type, then
    the operand with unsigned integer type is converted to the type of the
    operand with signed integer type. */
    else if (IS_SIGNED(lhs) && SIZEOF(lhs) > SIZEOF(rhs)) {
        post_sign_changed = cr;
        pre_sign_changed = r;
        cr->type = cl->type;
    }
    else if (IS_SIGNED(rhs) && SIZEOF(rhs) > SIZEOF(lhs)) {
        post_sign_changed = cl;
        pre_sign_changed = l;
        cl->type = cr->type;
    }
    /* Otherwise, both operands are converted to the unsigned integer type
    corresponding to the type of the operand with signed integer type. */
    else if (IS_SIGNED(lhs)) {
        post_sign_changed = cl;
        pre_sign_changed = l;
        cl->type = cr->type = MAKE_UNSIGNED(lhs);

    }
    else if (IS_SIGNED(rhs)) {
        post_sign_changed = cr;
        pre_sign_changed = r;
        cl->type = cr->type = MAKE_UNSIGNED(rhs);
    }
    if (post_sign_changed) {
        sign_extend(post_sign_changed,pre_sign_changed);
        if (sign(post_sign_changed) != sign(pre_sign_changed)) {
            char *orig = format_int(pre_sign_changed);
            char *converted = format_int(post_sign_changed);
            report(GRIPE_SIGN_CHANGED,NULL,"Integer %s changed to %s by the usual arithmetic conversions",
                   orig,converted);
            free(orig);
            free(converted);
        }
    }
}

/*! Check the validity of a shift operation. The operation has undefined
    behaviour if the promoted righthand operand is negative or is >= the
    width in bits of the lefthand operand.

    \param direction String specifying the direction of shift: "left" or "right".
    \param  type        The \e int_type of the lefthand op[erand.
    \param  shiftval    The integer value of the righthand operand.
*/
static bool
check_shift_validity(char const *direction, int_type type, int_spec_t const * shiftval)
{
    bool ok = true;
    unsigned type_bits = SIZEOF(type) * 8;
    unsigned long long val = shiftval->val.ull;
    if (IS_SIGNED(shiftval->type)) {
        unsigned long long neg_bit = 1 << (type_bits - 1);
        if (val & neg_bit) {
            report(GRIPE_NEGATIVE_SHIFT,NULL,"Negative %s-shift has undefined behavior. "
                   "The expression will not be resolved",direction);
            ok = false;
        }
    }
    if (ok && val >= type_bits) {
        report(GRIPE_SHIFT_OVERFLOW,NULL,"Shifting a quantity passed its width has undefined behavior. The expression will not be resolved");
        ok = false;
    }
    return ok;
}

/*@}*/

/* API*/

int_spec_t
lt(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = {INT_UNION_INITOR,INT_INT};
    int_spec_t cl,cr;
    do_usual_conversions(l,r,&cl,&cr);
    result.type = cl.type;
    switch(result.type) {
    case INT_INT:
        result.val.i = cl.val.i < cr.val.i;
        break;
    case INT_UINT:
        result.val.i = cl.val.ui < cr.val.ui;
        break;
    case INT_LONG:
        result.val.i = cl.val.l < cr.val.l;
        break;
    case INT_ULONG:
        result.val.i = cl.val.ul < cr.val.ul;
        break;
    case INT_LLONG:
        result.val.i = cl.val.ll < cr.val.ll;
        break;
    case INT_ULLONG:
        result.val.i = cl.val.ull < cr.val.ull;
        break;
    default:
        assert(false);
    }
    return result;
}

int_spec_t
gt(int_spec_t const * l, int_spec_t const *r)
{
    return lt(r,l);
}

int_spec_t
le(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = INT_SPEC_INITOR;
    int_spec_t cl,cr;
    do_usual_conversions(l,r,&cl,&cr);
    result.type = cl.type;
    switch(result.type) {
    case INT_INT:
        result.val.i = cl.val.i <= cr.val.i;
        break;
    case INT_UINT:
        result.val.i = cl.val.ui <= cr.val.ui;
        break;
    case INT_LONG:
        result.val.i = cl.val.l <= cr.val.l;
        break;
    case INT_ULONG:
        result.val.i = cl.val.ul <= cr.val.ul;
        break;
    case INT_LLONG:
        result.val.i = cl.val.ll <= cr.val.ll;
        break;
    case INT_ULLONG:
        result.val.i = cl.val.ull <= cr.val.ull;
        break;
    default:
        assert(false);
    }
    return result;
}

int_spec_t
ge(int_spec_t const * l, int_spec_t const *r)
{
    return le(r,l);
}

int_spec_t
eq(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = {INT_UNION_INITOR,INT_INT};
    result.val.i = l->val.ll == r->val.ll;
    return result;
}

int_spec_t
ne(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = {INT_UNION_INITOR,INT_INT};
    result.val.i = l->val.ll != r->val.ll;
    return result;

}

int_spec_t
bit_and(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = INT_SPEC_INITOR;
    int_spec_t cl,cr;
    do_usual_conversions(l,r,&cl,&cr);
    result.type = cl.type;
    result.val.ll = cl.val.ll & cr.val.ll;
    return result;
}

int_spec_t
bit_or(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = INT_SPEC_INITOR;
    int_spec_t cl,cr;
    do_usual_conversions(l,r,&cl,&cr);
    result.type = cl.type;
    result.val.ll = cl.val.ll | cr.val.ll;
    return result;
}

int_spec_t
bit_xor(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = INT_SPEC_INITOR;
    int_spec_t cl,cr;
    do_usual_conversions(l,r,&cl,&cr);
    result.type = cl.type;
    result.val.ll = cl.val.ll ^ cr.val.ll;
    return result;
}

int_spec_t
complement(int_spec_t const * int_spec)
{
    int_spec_t result = INT_SPEC_INITOR;
    result.type = int_spec->type;
    switch(result.type) {
    case INT_INT:
    case INT_UINT:
        result.val.ui = ~int_spec->val.ui;
        break;
    case INT_LONG:
    case INT_ULONG:
        result.val.ul = ~int_spec->val.ul;
        break;
    case INT_LLONG:
    case INT_ULLONG:
        result.val.ull = ~int_spec->val.ull;
    default:
        assert(false);
    }
    return result;
}

int_spec_t
minus(int_spec_t const * int_spec)
{
    int_spec_t result = INT_SPEC_INITOR;
    result.type = int_spec->type;
    switch(result.type) {
    case INT_INT:
        result.val.i = -int_spec->val.i;
        break;
    case INT_UINT:
        result.val.ui = -int_spec->val.ui;
        break;
    case INT_LONG:
        result.val.l = -int_spec->val.l;
        break;
    case INT_ULONG:
        result.val.ul = -int_spec->val.ul;
        break;
    case INT_LLONG:
        result.val.ll = -int_spec->val.ll;
        break;
    case INT_ULLONG:
        result.val.ull = -int_spec->val.ull;
        break;
    default:
        assert(false);
    }
    return result;
}

int_spec_t
lshift(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = INT_SPEC_INITOR;
    int_type type = l->type;
    bool valid_shift = check_shift_validity("left",type,r);
    if (valid_shift) {
        result.type = type;
        switch(type) {
        case INT_INT:
            result.val.i = l->val.i << r->val.i;
            break;
        case INT_UINT:
            result.val.ui = l->val.ui << r->val.i;
            break;
        case INT_LONG:
            result.val.l = l->val.l << r->val.i;
            break;
        case INT_ULONG:
            result.val.ul = l->val.ul << r->val.i;
            break;
        case INT_LLONG:
            result.val.ll = l->val.ll << r->val.i;
            break;
        case INT_ULLONG:
            result.val.ull = l->val.ull << r->val.i;
            break;
        default:
            assert(false);
        }
    }
    return result;
}

int_spec_t
rshift(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = INT_SPEC_INITOR;
    int_type type = l->type;
    bool valid_shift = check_shift_validity("left",type,r);
    if (valid_shift) {
        result.type = type;
        switch(type) {
        case INT_INT:
            result.val.i = l->val.i >> r->val.i;
            break;
        case INT_UINT:
            result.val.ui = l->val.ui >> r->val.i;
            break;
        case INT_LONG:
            result.val.l = l->val.l >> r->val.i;
            break;
        case INT_ULONG:
            result.val.ul = l->val.ul >> r->val.i;
            break;
        case INT_LLONG:
            result.val.ll = l->val.ll >> r->val.i;
            break;
        case INT_ULLONG:
            result.val.ull = l->val.ull >> r->val.i;
            break;
        default:
            assert(false);
        }
    }
    return result;
}

int_spec_t
add(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = INT_SPEC_INITOR;
    int_spec_t cl,cr;
    do_usual_conversions(l,r,&cl,&cr);
    result.type = cl.type;
    switch(result.type) {
    case INT_INT:
        result.val.i = cl.val.i + cr.val.i;
        break;
    case INT_UINT:
        result.val.ui = cl.val.ui + cr.val.ui;
        break;
    case INT_LONG:
        result.val.l = cl.val.l + cr.val.l;
        break;
    case INT_ULONG:
        result.val.ul = cl.val.ul + cr.val.ul;
        break;
    case INT_LLONG:
        result.val.ll = cl.val.ll + cr.val.ll;
        break;
    case INT_ULLONG:
        result.val.ull = cl.val.ull + cr.val.ull;
        break;
    default:
        assert(false);
    }
    return result;
}

int_spec_t
subtract(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = INT_SPEC_INITOR;
    int_spec_t cl,cr;
    do_usual_conversions(l,r,&cl,&cr);
    result.type = cl.type;
    switch(result.type) {
    case INT_INT:
        result.val.i = cl.val.i - cr.val.i;
        break;
    case INT_UINT:
        result.val.ui = cl.val.ui - cr.val.ui;
        break;
    case INT_LONG:
        result.val.l = cl.val.l - cr.val.l;
        break;
    case INT_ULONG:
        result.val.ul = cl.val.ul - cr.val.ul;
        break;
    case INT_LLONG:
        result.val.ll = cl.val.ll - cr.val.ll;
        break;
    case INT_ULLONG:
        result.val.ull = cl.val.ull - cr.val.ull;
        break;
    default:
        assert(false);
    }
    return result;
}

int_spec_t
multiply(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = INT_SPEC_INITOR;
    int_spec_t cl,cr;
    do_usual_conversions(l,r,&cl,&cr);
    result.type = cl.type;
    switch(result.type) {
    case INT_INT:
        result.val.i = cl.val.i * cr.val.i;
        break;
    case INT_UINT:
        result.val.ui = cl.val.ui * cr.val.ui;
        break;
    case INT_LONG:
        result.val.l = cl.val.l * cr.val.l;
        break;
    case INT_ULONG:
        result.val.ul = cl.val.ul * cr.val.ul;
        break;
    case INT_LLONG:
        result.val.ll = cl.val.ll * cr.val.ll;
        break;
    case INT_ULLONG:
        result.val.ull = cl.val.ull * cr.val.ull;
        break;
    default:
        assert(false);
    }
    return result;
}

int_spec_t
divide(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = INT_SPEC_INITOR;
    int_spec_t cl,cr;
    do_usual_conversions(l,r,&cl,&cr);
    result.type = cl.type;
    switch(result.type) {
    case INT_INT:
        result.val.i = cl.val.i / cr.val.i;
        break;
    case INT_UINT:
        result.val.ui = cl.val.ui / cr.val.ui;
        break;
    case INT_LONG:
        result.val.l = cl.val.l / cr.val.l;
        break;
    case INT_ULONG:
        result.val.ul = cl.val.ul / cr.val.ul;
        break;
    case INT_LLONG:
        result.val.ll = cl.val.ll / cr.val.ll;
        break;
    case INT_ULLONG:
        result.val.ull = cl.val.ull / cr.val.ull;
        break;
    default:
        assert(false);
    }
    return result;
}

int_spec_t
modulus(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = INT_SPEC_INITOR;
    int_spec_t cl,cr;
    do_usual_conversions(l,r,&cl,&cr);
    result.type = cl.type;
    switch(result.type) {
    case INT_INT:
        result.val.i = cl.val.i % cr.val.i;
        break;
    case INT_UINT:
        result.val.ui = cl.val.ui % cr.val.ui;
        break;
    case INT_LONG:
        result.val.l = cl.val.l % cr.val.l;
        break;
    case INT_ULONG:
        result.val.ul = cl.val.ul % cr.val.ul;
        break;
    case INT_LLONG:
        result.val.ll = cl.val.ll % cr.val.ll;
        break;
    case INT_ULLONG:
        result.val.ull = cl.val.ull % cr.val.ull;
        break;
    default:
        assert(false);
    }
    return result;
}

/*! boolean_and op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the conjunction as an \em int_spec_t
*/
int_spec_t
boolean_and(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = {INT_UNION_INITOR,INT_INT};
    result.val.i = l->val.ull && r->val.ull;
    return result;
}

/*! boolean_or op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the conjunction as an \em int_spec_t
*/
int_spec_t
boolean_or(int_spec_t const * l, int_spec_t const *r)
{
    int_spec_t result = {INT_UNION_INITOR,INT_INT};
    result.val.i = l->val.ull || r->val.ull;
    return result;
}


/*! boolean_negation op on an \em int_spec_t object.
    \param int_spec   The \em int_spec to be negated.
    \return The result of the negation as an \em int_spec_t
*/
int_spec_t
negate(int_spec_t const * int_spec)
{
    int_spec_t result = {INT_UNION_INITOR,INT_INT};
    result.val.i = !int_spec->val.ull;
    return result;
}


/* EOF*/


