#ifndef PTR_VECTOR_H
#define PTR_VECTOR_H

/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "bool.h"
#include <stddef.h>


/*! \ingroup ptr_vector_component ptr_vector_interface
    \file ptr_vector.h
    This file provides the Pointer-Vector component interface.
*/

/*!	\addtogroup ptr_vector_interface*/
/*@{*/

/*! Abstract signed type of a memory location or array offset*/
typedef ptrdiff_t locator_t;

/*! Abstract type of pointer-vector object*/
typedef struct ptr_vector * ptr_vector_h;

/*! Abstract type of const pointer-vector object*/
typedef struct ptr_vector const * ptr_vector_const_h;

/*! Abstact type of function for disposing of heap objects*/
typedef void (*dtor_t)(void *);

/*! Abstact type of function for cloning heap objects*/
typedef void * (*cloner_t)(void const *);

/*! Abstract type of function for comparing an object with a key value.
	\param	obj	Address of the object to be compared with the key.
	\param	key	Address of the key value.
	\param	keylen	If positive, this is the length of the key. If 0
			or negative then the function shall assume that \e key is in fact
			the address of a second object of the same type as \e obj and shall
			compare thse two objects by key.
	\return < 0 of \e obj compares less than \e key; 0 if \e obj compares
			equal with \e key; > 0 if \e obj compares greater than \e key.
*/
typedef int (*comparator_t)(void const *obj, void const *key, locator_t keylen);

/*! Predicate denoting that a location exists */
#define SOMEWHERE(loc)	((loc) >= 0)

/*! Predicate denoting a non-existent location*/
#define NOWHERE(loc)	((loc) < 0)

/*! Get the pointer represented by a locator_t */
#define POINTER(loc) (NOWHERE(loc) ? ~(loc) : (loc))

/*! Create new pointer-vector object
	\param	dtor	The destructor to be applied to vector
					elements when the vector is destroyed, or NULL.
					If NULL then \e free() is assumed.
	\param	cloner	The cloning function for cloning elements of this
					vector, or NULL. If NULL then \e clone(ptr,0) is
					assumed.
    \return A new \c ptr_vector object
*/
extern ptr_vector_h
ptr_vector_new(dtor_t dtor, cloner_t cloner);

/*! Dispose of a pointer-vector object
    \param pv The pointer-vector to be destroyed.

	The vector's \e dtor member is called on each element
    and then remaining memory controlled by the pointer-vector is
    released.

	The function is a no-op if \e pv is NULL.
*/
extern void
ptr_vector_dispose(ptr_vector_h pv);

/*! Construct a copy of a ptr_vector.
	\param	src	The ptr_vector to copy
	\return A copy of \e src.
*/
extern ptr_vector_h
ptr_vector_copy(ptr_vector_const_h src);

/*! Exchange two pointer vectors*/
extern void
ptr_vector_swap(ptr_vector_h lhs, ptr_vector_h rhs);

/*! Assign one pointer vector to another.
	\param	dest	The pointer vector to be assigned
	\param	src		The pointrer vector to be assigned to \e dest.
*/
extern void
ptr_vector_assign(ptr_vector_h dest, ptr_vector_const_h src);

/*! Insert a pointer into a pointer-vector.
    \param pv	The pointer-vector in which a pointer is to be inserted.
    \param	pos	The position in the vector at which to insert.
    \param ptr	The pointer to be inserted.
    If \c pos is beyond the end of \c pv then \c ptr is appended to \c pv
    The vector is extended if necessary to insert or append \c ptr.
    \note Pointers that the caller is holding to pointers within the vector
    must be assumed invalidated after an insertion.
    \note Insertion is inefficient even when it does not provoke an
    extension of the vector, because it requires all elements from \c pos
    onward to be moved up 1 position.
*/
extern void
ptr_vector_insert(ptr_vector_h pv, size_t pos, void const *ptr);


/*! Append a pointer to a pointer-vector.
    \param pv	The pointer-vector to which the pointer is to be appended.
    \param ptr	The pointer to be appended.
    The vector is extended if necessary to append \c ptr.
    \note Pointers that the caller is holding to pointers within the vector
    must be assumed invalidated after an insertion.
*/
extern void
ptr_vector_append(ptr_vector_h pv, void *ptr);

/*!	Get the destructor function from a pointer vector.
*/
extern dtor_t
ptr_vector_dtor(ptr_vector_const_h pv);

/*!	Get the cloning function from a pointer vector.
*/
extern cloner_t
ptr_vector_cloner(ptr_vector_const_h pv);

/*! Get the number of elements in a pointer-vector.
    \param pv	The pointer-vector to count.
    \return 	The number of elements in \c pv.
*/
extern size_t
ptr_vector_count(ptr_vector_const_h pv);

/*! Say whether two pointer vectors compare equal.
	\param	lhs	The first pointer vector to compare.
	\param	rhs	The second pointer vector to compare.
	\param	compare	The comparator function to be used.
	\return True if both vectors are NULL or if they have:
			- the same destructor and cloner members
			- the same number of elements that are pairwise equal
				as evaluated by	\e compare.
*/
extern bool
ptr_vector_equal(	ptr_vector_const_h lhs,
                    ptr_vector_const_h rhs,
                    comparator_t compare);

/*! Get a pointer to the start of a pointer-vector.
    \param pv	The pointer-vector whose start is wanted.
    \return 	Pointer to the first element of \c pv, if
    			\c pv is non-empty; otherwise \c NULL
*/
extern void **
ptr_vector_begin(ptr_vector_h pv);

/*! Get a const pointer to the start of a pointer-vector.
    \param pv	The pointer-vector whose start is wanted.
    \return 	Pointer to the first element of \c pv, if
    			\c pv is non-empty; otherwise \c NULL
*/
extern void const **
ptr_vector_begin_const(ptr_vector_const_h pv);

/*! Get a pointer to the end of a pointer-vector.
    \param pv	The pointer-vector whose end is wanted.
    \return 	Pointer just past the last element of \c pv, if
    			\c pv is non-empty; otherwise \c NULL
*/
extern void **
ptr_vector_end(ptr_vector_h pv);

/*! Get a const pointer to the end of a pointer-vector.
    \param pv	The pointer-vector whose end is wanted.
    \return 	Pointer just past the last element of \c pv, if
    			\c pv is non-empty; otherwise \c NULL
*/
extern void const **
ptr_vector_end_const(ptr_vector_const_h pv);


/*! Get a pointer to the element at a given position in a pointer-vector.
    \param pv	The pointer-vector from which an element is wanted.
    \param pos	The position in the vector from which the element is wanted.
    \return 	The pointer at position \c pos in  \c pv, unless \c pos
    			is out of range; \c NULL if \c pos is out of range.
*/
extern void *
ptr_vector_at(ptr_vector_h pv, size_t pos);

/*! Get a const pointer to the element at a given position in a pointer-vector.
    \param pv	The pointer-vector from which an element is wanted.
    \param pos	The position in the vector from which the element is wanted.
    \return 	The pointer at position \c pos in  \c pv, unless \c pos
    			is out of range; \c NULL if \c pos is out of range.
*/
extern void const *
ptr_vector_at_const(ptr_vector_const_h pv, size_t pos);

/*! Search a pointer vector for an element matching a key value
	\param	pv	The pointer vector to search
	\param	key	The address of the key value.
	\param	keylen	The Length of the key value or 0.
	\param	compare	The comparator function to use for matching.
	\param	ordered	If true then elements of the \e pv are
		assumed to be ordered ascending by \e compare. In this case
		the search can be accelerated.
	\return	If \e key is matched then the index of the matching element
	is returned, otherwise a negative value L. If \e ordered is
	true then ~L is index at which at element matching \e key should
	be inserted to preserve order.
*/
extern locator_t
ptr_vector_search(	ptr_vector_const_h pv,
                    void const *key, locator_t keylen, comparator_t compare,
                    bool ordered);

/*@}*/

#endif /* EOF*/
