#ifndef SYMBOL_H
#define SYMBOL_H
/***************************************************************************
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "bool.h"
#include "canonical_string.h"
#include "eval_result.h"
#include <math.h>
#include <stddef.h>
#include <ctype.h>

/*!\ingroup symbol_component symbol_interface
 *\file symbol.h
 * This file provides the Symbol component interface
*/

/*!\addtogroup symbol_interface*/
/*@{*/

/*! Abstract type of symbol object*/
typedef struct eval_result * symbol_h;

/*! Abstract type of const symbol object*/
typedef struct eval_result const * symbol_const_h;

/*! Test whether a symbol is defined.
    \param sym	 The symbol to test.
*/
#define SYM_DEFINED(sym)	(symbol_def_len(sym) >= 0)

/*! Construct a new symbol object
	\param	name	Pointer to the name of the symbol.
	\param namelen	The length of the symbol name.
	\return A symbol object having the specified name, no definition and
		0 tally.
*/
extern symbol_h
symbol_new(char const *name, size_t namelen);

/*! Dispose of a symbol object, releasing its resources.
	If \e sym is NULL the call is a no-op.
*/
extern void
symbol_dispose(symbol_h sym);

/*! Construct a copy of a symbol.
	\param	src	The symbol to copy
	\return A copy of \e src.
*/
extern symbol_h
symbol_copy(symbol_const_h src);

/*! Exchange two symbols*/
extern void
symbol_swap(symbol_h lhs, symbol_h rhs);

/*! Assign one symbol to another.
	\param	dest	The symbol to be assigned
	\param	src		The symbol to be assigned to \e dest.
*/
extern void
symbol_assign(symbol_h dest, symbol_const_h src);

/*! Say whether two symbols are equal.
	\param	lhs	The first symbol to compare.
	\param	rhs	The second symbol to compare.
	\return True if both symbols are NULL or if they have the same
		name and are identically defined or undefined.
*/
extern bool
symbol_equal(symbol_const_h lhs,symbol_const_h rhs);


/*! Assign a definition to a symbol.
	\param	sym	The symbol to be defined.
	\param	pdef	NULL, or a pointer to the start of the 0-terminated text containing the
        symbol definition, possibly polluted with comments and a terminal linefeed.
        If not NULL, then on return \e pdef stores the address that terminates
        this text. If \e def is NULL the symbol is defined	as an empty string.
*/
extern void
symbol_define(symbol_h sym, char const **pdef);

/*! Remove any definition from a symbol.
	\param	sym	The symbol to be undefined.
*/
extern void
symbol_undefine(symbol_h sym);

/*! Get the name of a symbol.
	\param	sym	The symbol whose name is to be queried.
	\return	Pointer to the 0-terminated name of the symbol.
*/
extern char const *
symbol_name(symbol_const_h sym);

/*! Get the definition of a symbol.
	\param	sym	The symbol whose definition is to be queried.
	\return	A canonical string containing the definition of \e sym,
		if the symbol is defined, otherwise NULL.
*/
extern canonical_string_const_h
symbol_definition(symbol_const_h sym);

/*! Get the final value of a symbol.
	\param	sym	The symbol whose final value is to be queried.
	\return	A canonical string containing the final value of \e sym,
		if the symbol is defined, otherwise NULL.

	The function recursively retrieves symbol definitions as long
	as the definition of the symbol in hand is itself a defined symbol
	and returns the first definition that is not a defined symbol.
*/
extern canonical_string_const_h
symbol_final_value(symbol_h sym);


/*! Get the length of a symbol definition
	\param	sym	The symbol whose definition is to be queried.
	\return	-1 if the symbol is undefined; 0 if the symbol has an
	an empty definition; otherwise the length of the symbol definition.
*/
extern int
symbol_def_len(symbol_const_h sym);

/*! Mark a symbol as seen
	\param	sym	The symbol to memorise.
*/
extern void
symbol_memorise(symbol_h sym);

/*! Say whether a symbol has been seen.
	\param	sym	The symbol to be queried.
	\return	True if the symbol has been memorised and not subsequently
			forgotten.
*/
extern bool
symbol_seen(symbol_const_h sym);

/*! Forget that a symbol has been seen.
	\param	sym	The symbol to be forgotten
*/
extern void
symbol_forget(symbol_h sym);

/*! Compare a symbol with another symbol or a name.
	\param	lhs	First symbol to be compared.
	\param	rhs	Second symbol or name to compare.
	\param	namelen	If 0 then \e rhs addresses a second symbol to be
			compared by name with \e lhs. Otherwise \e rhs addresses a name
			of length \e namelen to be compared with the name of \e lhs.
	\return A negative value if \e lhs is less than \e rhs; 0 if \e lhs
		is equal to \e rhs; a positive value if \e lhs is greater than \e rhs.
*/
extern int
symbol_compare(symbol_const_h lhs, void const * rhs, size_t namelen);

/*! Say whether a token is legal preprocesor symbol name.
	\param	tok		Start of the token.
	\param	toklen	Length of the token
	\return	True if the token comprises a legal preprocessor symbol name.
*/
extern bool
symbol_is_legal(char const *tok, size_t toklen);

/*! Say whether a character can be the first
	character of a symbol.
*/
extern bool
is_symbol_start_char(char cp);

/*! Say whether can occur within a symbol,
	except possibly in the first position.
*/
extern bool
is_symbol_inner_char(char cp);

/*!	Resolve a symbol.

	\param sym	The symbol to resolved.

	If \e sym has already been seen and resolved
	the function is a no-op.

	Otherwise \e sym is resolved by:-

	- Determining and assigning the integer value, if any, that accrues
		to \e sym
	- Constructing and assigning the most-resolved form that accrues to \e sym

	\e sym will acquire an integer value if is defined as a soluble
	legal expression.

	\e sym always acquires a most-resolved form:-

	- If \e sym is configured with \c --define \c SYM then
		its most-resolved form is the empty string.

	- If \e sym is configured with \c --undef \c SYM then
		its most-resolved form is "0"

	- If \e sym is configured with \c --def \c SYM=EXPR.
	where \c EXPR is a soluble expression with integer value \e N,
	then the most-resolved of \e sym is the numeral denoting \e N.

	- If \e sym is configured with \c --def \c SYM=VAL.
	where \c VAL is an arbitrary string, then the most-resolved form of
	\e sym results from \c VAL by:-

	- Obtaining the canonical form of \c VAL
	- Replacing from left to right each maximally long legal expression
	in the canonical form of \c VAL with its most resolved form and
	leaving the other tokens unchanged.

*/
extern void
symbol_resolve(symbol_h sym);

/*@}*/

#endif /* EOF*/
